#
#   5B - Central game state
#

import os, sys
from tempfile import mkstemp
from glob import glob
from albow.dialogs import ask
from albow.file_dialogs import look_for_file_or_directory
from level import Level
from reader import Reader
from settings import options
import settings, version

class Game(object):
	#  current_level          Level
	#  saved_game_pathname    string
	#  completed_levels       [string]   level file basenames
	#  unsaved_progress       boolean
	#  current_level_dir      string
	
	game_file_suffix = ".5bg"
	game_file_magic = "5BG"
	game_file_version = 2
	
	level_set_suffix = ".5bs"
	
	def __init__(self):
		#  ALL instance variables must be initialised here and not as
		#  class variables, so that game load/save works correctly.
		self.current_level = None
		self.current_level_dir = None
		self.current_save_dir = None
		self.clear_progress()

	def start_enabled(self):
		return self.level_present() and not self.level_running()
		
	def resume_enabled(self):
		return self.level_running()

	def abort_enabled(self):
		return self.level_running()
		
	def save_game_enabled(self):
		return bool(self.current_level_dir)
	
	def std_levels_enabled(self):
		return self.playing_custom_levels()
	
	def playing_custom_levels(self):
		cur_dir = self.current_level_dir
		std_dir = settings.get_default_playing_level_dir()
		result = cur_dir and cur_dir <> std_dir
		return result

	def level_present(self):
		return self.current_level is not None
	
	def level_running(self):
		level = self.current_level
		if level:
			return level.playing()
		else:
			return False
	
	def start_level(self, testing = False):
		level = self.current_level
		if level:
			if testing:
				level.testing = True
			level.restart()
	
	def abort_level(self):
		level = self.current_level
		if level:
			level.reset()
	
	def set_current_level(self, level):
		level.game = self
		self.current_level = level
		path = level.pathname
		if path:
			level_dir = os.path.dirname(path)
			self.set_current_level_dir(level_dir)

	def create_level(self):
		self.set_current_level(Level())

	def load_level(self, path):
		f = open(path, "rU")
		try:
			r = Reader(f, path)
			level = Level.create_from_file(r)
		finally:
			f.close()
		level.is_tutorial = level.get_name().startswith("+")
		self.set_current_level(level)
	
	def load_game(self, path):
		f = open(path, "rU")
		try:
			return self.read_game_from_file(f)
		finally:
			f.close()

	def read_game_from_file(self, f):
		def xint(s):
			try:
				return int(s)
			except ValueError:
				return sys.maxint
		header = f.readline()
		parts = header.split()
		if len(parts) < 2 or parts[0] <> self.game_file_magic \
			or (len(parts) < 3 and xint(parts[1]) >= 2):
			raise EnvironmentError("Does not seem to be a saved game file.")
		if len(parts) >= 3:
			creator = parts[2]
		else:
			creator = "unknown"
		if xint(parts[1]) > self.game_file_version:
			raise EnvironmentError("Can't read this game file. "
				"It was created by 555-BOOM! version %s." % creator)
		line = f.readline()
		if not line:
			raise EnvironmentError("Unexpected end of file.")
		c_level_dir = canonicalize_path(line.strip())
		completed = []
		for line in f.readlines():
			completed.append(line.strip())
		if c_level_dir == "Resources/levels":
			#level_dir = os.path.join("Game", level_dir)
			c_level_dir = "Game/" + c_level_dir
		level_dir = decanonicalize_path(c_level_dir)
		program_dir = settings.get_program_dir()
		level_dir = os.path.join(program_dir, level_dir)
		level_dir = self.validate_level_dir(level_dir)
		if level_dir:
			self.current_level_dir = level_dir
			self.completed_levels = completed
			return True
		else:
			return False
	
	def validate_level_dir(self, path):
		if os.path.exists(path):
			return path
		elif ask("Can't find '%s'. Would you like to look for it?" % path,
			["OK", "Cancel"]) <> "OK":
				return None
		else:
			name = os.path.basename(path)
			program_dir = settings.get_program_dir()
			return look_for_file_or_directory(target = name,
				directory = program_dir)
	
	def create_level_set(self, path):
		create_containing_directory(path)
		os.mkdir(path)

	def save_level(self):
		level = self.current_level
		path = level.pathname
		create_containing_directory(path)
		fd, temp_path = mkstemp(dir = os.path.dirname(path), text = True)
		f = os.fdopen(fd, "w")
		try:
			try:
				level.write_to_file(f)
			finally:
				f.close()
		except:
			os.unlink(temp_path)
			raise
		else:
			os.rename(temp_path, path)
			level.changed = False
	
	def save_game(self, path):
		create_containing_directory(path)
		fd, temp_path = mkstemp(dir = os.path.dirname(path), text = True)
		f = os.fdopen(fd, "w")
		try:
			try:
				self.write_game_to_file(f)
			finally:
				f.close()
		except:
			os.unlink(temp_path)
			raise
		else:
			os.rename(temp_path, path)
		self.unsaved_progress = False
	
	def write_game_to_file(self, f):
		level_dir = canonicalize_path(self.current_level_dir)
		program_dir = canonicalize_path(settings.get_program_dir())
		level_dir = strip_canonical_path_prefix(program_dir, level_dir)
		f.write("%s %s %s\n" % (self.game_file_magic, self.game_file_version,
			version.version))
		f.write("%s\n" % level_dir)
		for name in self.completed_levels:
			f.write("%s\n" % name)

	def get_current_playing_level_list(self, tutorial = True, regular = True):
		def filter(path):
			is_tutorial = os.path.basename(p).startswith("+")
			if is_tutorial and not tutorial:
				return False
			if not is_tutorial and not regular:
				return False
			return True
		level_dir = self.get_current_playing_level_dir()
		pat = os.path.join(level_dir, "*%s" % Level.level_file_suffix)
		paths = glob(pat)
		paths = [p for p in paths if filter(p)]
		paths.sort() # Glob results not sorted on some systems, it seems
		return paths

	def get_current_playing_level_dir(self):
		return self.current_level_dir \
			or settings.get_default_playing_level_dir()

	def get_adjacent_level_path(self, way):
		paths = self.get_current_playing_level_list()
		try:
			i = paths.index(self.current_level.pathname)
		except ValueError:
			return None
		i += way
		if 0 <= i < len(paths):
			return paths[i]
		
	def get_current_save_dir(self):
		path = self.saved_game_pathname
		if path:
			return os.path.dirname(path)
		else:
			return settings.get_default_save_dir()
	
	def get_current_custom_level_dir(self):
		old_dir = self.current_level_dir
		#print "Game.get_current_custom_level_dir: current_level_dir =", old_dir ###
		std_dir = settings.get_default_playing_level_dir()
		if not old_dir:
			old_dir = settings.get_default_editor_level_dir()
		elif old_dir == std_dir and not options.edit_std_levels:
			old_dir = settings.get_default_editor_level_dir()
		return old_dir

	def set_current_level_dir(self, directory):
		if self.current_level_dir <> directory:
			self.clear_progress()
		self.current_level_dir = directory

	def clear_progress(self):
		self.completed_levels = []
		self.saved_game_pathname = None
		self.unsaved_progress = False

	def record_progress(self, level_name):
		if level_name and level_name not in self.completed_levels:
			self.completed_levels.append(level_name)
	
	def next_uncompleted_level(self):
		level_dir = self.get_current_playing_level_dir()
		completed = self.completed_levels
		paths = self.get_current_playing_level_list(tutorial = False)
		names = [os.path.basename(p) for p in paths]
		names.sort()
		completed.sort()
		i = 0
		n = min(len(names), len(completed))
		while i < n and names[i] == completed[i]:
			i += 1
		if i < len(names):
			return os.path.join(level_dir, names[i])
		else:
			return None

#def strip_path_prefix(prefix, path):
#	#print "strip_path_prefix:", prefix, path ###
#	result = []
#	dir = path
#	while 1:
#		dir2, name = os.path.split(dir)
#		#print "dir =", dir, "name =", name, "result =", result ###
#		if dir2 == dir:
#			#print "prefix not found" ###
#			return path
#		result.insert(0, name)
#		if dir2 == prefix:
#			#print "prefix found" ###
#			return os.path.join(*result)
#		dir = dir2

def filename_to_level_name(filename):
	prefix = ""
	name = os.path.splitext(filename)[0]
	if name.startswith("+"):
		prefix = "Tutorial "
		name = name[1:]
	name = name.replace("-", ": ")
	name = name.replace("_", " ")
	while name.startswith("0"):
		name = name[1:]
	return prefix + name

def canonicalize_path(path):
	return os.path.normpath(path).replace("\\", "/")

def decanonicalize_path(path):
	return os.path.join(*path.split("/"))

def strip_canonical_path_prefix(prefix, path):
	#  Prefix and path both assumed to be absolute paths.
	#  If prefix is stripped, result is a relative path.
	if not prefix.endswith("/"):
		prefix += "/"
	if not path.endswith("/"):
		path += "/"
	n = len(prefix)
	if path.startswith(prefix):
		path = path[n:]
		if path.endswith("/"):
			path = path[:-1]
	return path

def create_containing_directory(path):
	directory = os.path.dirname(path)
	if not os.path.exists(directory):
		os.makedirs(directory)
